@@ -1,14 +1,14 @@
 # -*- perl -*-
 
 #
-# $Id: Getopt.pm,v 1.45 2002/08/02 12:55:32 eserte Exp $
+# $Id: Getopt.pm,v 1.48 2003/09/16 11:57:19 eserte Exp eserte $
 # Author: Slaven Rezic
 #
-# Copyright (C) 1997,1998,1999,2000 Slaven Rezic. All rights reserved.
+# Copyright (C) 1997,1998,1999,2000,2003 Slaven Rezic. All rights reserved.
 # This package is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 #
-# Mail: slaven.rezic@berlin.de
+# Mail: slaven@rezic.de
 # WWW:  http://user.cs.tu-berlin.de/~eserte/
 #
 
@@ -24,7 +24,7 @@ use constant OPTTYPE  => 1;
 use constant DEFVAL   => 2;
 use constant OPTEXTRA => 3;
 
-$VERSION = '0.48';
+$VERSION = '0.49';
 
 $DEBUG = 0;
 $x11_pass_through = 0;
@@ -45,6 +45,9 @@ sub new {
 	    if (ref $_ eq 'ARRAY' and
 		defined $_->[OPTEXTRA] and
 		ref $_->[OPTEXTRA] ne 'HASH') {
+		if ((@$_ - OPTEXTRA) % 2 != 0) {
+		    warn "Odd number of elements in definition for " . $_->[OPTNAME];
+		}
 		my %h = splice @$_, OPTEXTRA;
 		$_->[OPTEXTRA] = \%h;
 	    }
@@ -184,7 +187,7 @@ sub set_defaults {
 	    } elsif ($ref eq 'SCALAR') {
 		$ {$self->_varref($opt)} = $opt->[DEFVAL];
 	    } else {
-		die "Invalid reference type for option $opt->[OPTNAME]";
+		die "Invalid reference type for option $opt->[OPTNAME] while setting the default value (maybe you should specify <undef> as the default value)";
 	    }
 	}
     }
@@ -243,11 +246,17 @@ sub save_options {
 	    my $opt;
 	    foreach $opt ($self->_opt_array) {
 		if (!$opt->[OPTEXTRA]{'nosave'}) {
-		    if (ref($self->_varref($opt)) eq 'SCALAR') {
-			$saveoptions{$opt->[OPTNAME]} = $ {$self->_varref($opt)}
-		    } elsif (ref($self->_varref($opt)) =~ /^(HASH|ARRAY)$/) {
-			$saveoptions{$opt->[OPTNAME]} = $self->_varref($opt);
-		    } 
+		    my $ref;
+		    if ($opt->[OPTEXTRA]{'savevar'}) {
+			$ref = $opt->[OPTEXTRA]{'savevar'};
+		    } else {
+			$ref = $self->_varref($opt);
+		    }
+		    if (ref($ref) eq 'SCALAR') {
+			$saveoptions{$opt->[OPTNAME]} = $$ref;
+		    } elsif (ref($ref) =~ /^(HASH|ARRAY)$/) {
+			$saveoptions{$opt->[OPTNAME]} = $ref;
+		    }
 		}
 	    }
 	    if (Data::Dumper->can('Dumpxs')) {
@@ -394,11 +403,12 @@ sub process_options {
 	    if (!grep($_ eq $v, @choices)) {
 		if (defined $former) {
 		    warn "Not allowed: " . $ {$self->_varref($_)}
-		    . " for $opt. Using old value $former->{$opt}";
+		    . " for -$opt. Using old value $former->{$opt}";
 		    $ {$self->_varref($_)} = $former->{$opt};
 		} else {
 		    die "Not allowed: "
-		      . $ {$self->_varref($_)} . " for $opt";
+		      . $ {$self->_varref($_)} . " for -$opt\n"
+		      . "Allowed is only: " . join(", ", @choices);
 		}
 	    }
 	}
@@ -409,7 +419,7 @@ sub process_options {
 sub _fix_layout {
     my($self, $frame, $widget, %args) = @_;
     my($w, $real_w);
-    if ($Tk::VERSION < 999) { # XXX
+    if ($Tk::VERSION < 804) {
 	my $f = $frame->Frame;
 	$f->Label->pack(-side => "left"); # dummy
 	$real_w = $f->$widget(%args)->pack(-side => "left", -padx => 1);
@@ -617,8 +627,12 @@ sub _filedialog_widget {
 	   if ($Tk::VERSION >= 800) {
 	       if ($subtype eq 'dir') {
 		   $fd = '_dir_select';
-	       } else {
+	       } elsif ($subtype eq 'savefile') {
+		   $fd = 'getSaveFile';
+	       } elsif ($subtype eq 'file') {
 		   $fd = 'getOpenFile';
+	       } else {
+		   die "Unknown subtype <$subtype>";
 	       }
 	   } else {
 	       $fd = 'FileDialog';
@@ -645,13 +659,13 @@ sub _filedialog_widget {
 	       $base = File::Basename::basename($act_val);
 	       $dir = '.' if (!-d $dir);
 
-	       if ($fd eq 'getOpenFile') {
-		   $file = $topframe->getOpenFile(-initialdir => $dir,
-						  -initialfile => $base,
-						  -title => 'Select file',
+	       if ($fd =~ /^get(Open|Save)File$/) {
+		   $file = $topframe->$fd(-initialdir => $dir,
+					  -initialfile => $base,
+					  -title => 'Select file',
 # XXX erst ab 800.013 (?)
 #						  -force => 1,
-						 );
+					 );
 	       } elsif ($fd eq '_dir_select') {
 		   $file = _dir_select($topframe, $dir);
 	       } elsif ($fd eq 'FileDialog') {
@@ -667,8 +681,8 @@ sub _filedialog_widget {
 		   }
 	       }
 	   } else {
-	       if ($fd eq 'getOpenFile') {
-		   $file = $topframe->getOpenFile(-title => 'Select file');
+	       if ($fd =~ /^get(Open|Save)File$/) {
+		   $file = $topframe->$fd(-title => 'Select file');
 	       } elsif ($fd eq '_dir_select') {
 		   require Cwd;
 		   $file = _dir_select($topframe, Cwd::cwd());
@@ -825,10 +839,10 @@ sub _create_page {
 	    my $subtype = (defined $opt->[OPTEXTRA] &&
 			   exists $opt->[OPTEXTRA]{'subtype'} ?
 			   $opt->[OPTEXTRA]{'subtype'} : "");
-	    if ($subtype eq 'file') {
-		$w = $self->_filedialog_widget($f, $opt);
-	    } elsif ($subtype eq 'dir') {
-		$w = $self->_filedialog_widget($f, $opt, -subtype => "dir");
+	    if ($subtype eq 'file' ||
+		$subtype eq 'savefile' ||
+		$subtype eq 'dir') {
+		$w = $self->_filedialog_widget($f, $opt, -subtype => $subtype);
 	    } elsif ($subtype eq 'geometry') {
 		$w = $self->_geometry_widget($f, $opt);
 	    } elsif ($subtype eq 'color') {
@@ -912,15 +926,16 @@ sub option_editor {
 			     ? delete $a{'-delaypagecreate'}
 			     : 1);
     if (!defined $string) {
-	$string = {'optedit'   => 'Option editor',
-		   'undo'      => 'Undo',
-		   'lastsaved' => 'Last saved',
-		   'save'      => 'Save',
-		   'defaults'  => 'Defaults',
-		   'ok'        => 'OK',
-		   'apply'     => 'Apply',
-		   'cancel'    => 'Cancel',
-		   'helpfor'   => 'Help for:',
+	$string = {'optedit'    => 'Option editor',
+		   'undo'       => 'Undo',
+		   'lastsaved'  => 'Last saved',
+		   'save'       => 'Save',
+		   'defaults'   => 'Defaults',
+		   'ok'         => 'OK',
+		   'apply'      => 'Apply',
+		   'cancel'     => 'Cancel',
+		   'helpfor'    => 'Help for:',
+		   'oksave'     => 'OK',
 	          };
     }
     $self->{_string} = $string;
@@ -1071,11 +1086,8 @@ sub option_editor {
         }
     }
 
-    my($ok_button, $apply_button, $cancel_button, $undo_button,
-       $lastsaved_button, $save_button, $def_button);
-
     if (!$buttons || $allowed_button{'ok'}) {
-	$ok_button
+	my $ok_button
 	    = $f->Button(-text => $string->{'ok'},
 			 -underline => 0,
 			 -command => sub {
@@ -1089,8 +1101,32 @@ sub option_editor {
         push @tiler_b, $ok_button;
     }
 
+    if ($allowed_button{'oksave'}) {
+	my $ok_button
+	    = $f->Button(-text => $string->{'oksave'},
+			 -underline => 0,
+			 -command => sub {
+			     $top->Busy;
+			     eval {
+				 $self->save_options;
+				 $self->process_options(\%undo_options, 1);
+				 if (!$dont_use_notebook) {
+				     $self->{'raised'} = $opt_notebook->raised();
+				 }
+			     };
+			     my $err = $@;
+			     $top->Unbusy;
+			     if ($err) {
+				 die $err;
+			     }
+                             $opt_editor->destroy;
+                         }
+                        );
+        push @tiler_b, $ok_button;
+    }
+
     if (!$buttons || $allowed_button{'apply'}) {
-	$apply_button
+	my $apply_button
 	    = $f->Button(-text => $string->{'apply'},
 			 -command => sub {
 			     $self->process_options(\%undo_options, 1);
@@ -1098,7 +1134,8 @@ sub option_editor {
 			);
 	push @tiler_b, $apply_button;
     }
-
+	
+    my $cancel_button;
     if (!$buttons || $allowed_button{'cancel'}) {
 	$cancel_button
 	    = $f->Button(-text => $string->{'cancel'},
@@ -1114,7 +1151,7 @@ sub option_editor {
     }
 
     if (!$buttons || $allowed_button{'undo'}) {
-	$undo_button
+	my $undo_button
 	    = $f->Button(-text => $string->{'undo'},
 			 -command => sub {
 			     $self->_do_undo(\%undo_options);
@@ -1125,7 +1162,7 @@ sub option_editor {
 
     if ($self->{'filename'}) {
 	if (!$buttons || $allowed_button{'lastsaved'}) {
-	    $lastsaved_button
+	    my $lastsaved_button
 		= $f->Button(-text => $string->{'lastsaved'},
 			     -command => sub {
 				 $top->Busy;
@@ -1137,6 +1174,7 @@ sub option_editor {
 	}
 
 	if (!$nosave && (!$buttons || $allowed_button{'save'})) {
+	    my $save_button;
 	    $save_button
 		= $f->Button(-text => $string->{'save'},
 			     -command => sub {
@@ -1153,7 +1191,7 @@ sub option_editor {
     }
 
     if (!$buttons || $allowed_button{'defaults'}) {
-	$def_button
+	my $def_button
 	    = $f->Button(-text => $string->{'defaults'},
 			 -command => sub {
 			     $self->set_defaults;
@@ -1179,6 +1217,7 @@ sub option_editor {
     }
 
     if ($opt_editor->can('Popup')) {
+	$opt_editor->withdraw;
 	$opt_editor->Popup;
     }
     if ($wait) {
@@ -1468,6 +1507,12 @@ a reference to the option editor window.
 
 Disable saving of options.
 
+=item -savevar
+
+When saving with the C<saveoptions> method, use the specified variable
+reference instead of the C<-var> reference. This is useful if C<-var>
+is a subroutine reference.
+
 =item -buttons
 
 Specify, which buttons should be drawn. It is advisable to draw at
@@ -1623,12 +1668,14 @@ match either the choices or the range.
 
 =item subtype
 
-The subtypes are I<file>, I<dir>, I<geometry>, I<font> and I<color>.
-These can be used with string options. For the first two, the GUI
-interface will pop up a file dialog for this option (either for
-selecting files or directories). If the I<geometry> subtype is
-specified, the user can set the current geometry of the main window.
-The I<color> subtype is not yet implemented.
+Valid subtypes are I<file>, I<savefile>, I<dir>, I<geometry>, I<font>
+and I<color>. These can be used with string options. For I<file> and
+I<savefile>, the GUI interface will pop up a file dialog, using
+C<getOpenFile> for the former and C<getSaveFile> for the latter. For
+I<dir>, the GUI interface will pop up dialog for selecting
+directories. If the I<geometry> subtype is specified, the user can set
+the current geometry of the main window. The I<color> subtype is not
+yet implemented.
 
 =item var
 
@@ -1779,7 +1826,7 @@ This should be done only by Apply and Ok buttons.
 
 =head1 AUTHOR
 
-Slaven Rezic <slaven.rezic@berlin.de>
+Slaven Rezic <slaven@rezic.de>
 
 This package is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
